"""
Asynchronous client.

Abstract base class for other asynchronous clients such as WebSockets, ZeroMQ
and aiohttp.
"""
from abc import ABCMeta, abstractmethod

from future.utils import with_metaclass

from .client import Client
from .request import Request, Notification
from .prepared_request import PreparedRequest


class AsyncClient(with_metaclass(ABCMeta, Client)):
    @abstractmethod
    async def send_message(self, request, **kwargs):
        """(Abstract)"""

    async def send(self, request, **kwargs):
        request = PreparedRequest(request)
        self.prepare_request(request, **kwargs)
        self.log_request(request, request.log_extra, request.log_format)
        return await self.send_message(request, **kwargs)

    async def notify(self, method_name, *args, **kwargs):
        return await self.send(Notification(method_name, *args, **kwargs))

    async def request(self, method_name, *args, **kwargs):
        return await self.send(Request(method_name, *args, **kwargs))



"""
An HTTP client.

For example::

    HTTPClient('http://example.com/api').request('go')

Uses the `Requests <http://docs.python-requests.org/en/master/>`_ library.
"""
from requests import Request, Session

from .client import Client


class HTTPClient(Client):
    """Defines an HTTP client"""

    # The default HTTP header
    DEFAULT_HEADERS = {
        'Content-Type': 'application/json', 'Accept': 'application/json'}

    def __init__(self, endpoint):
        """
        :param endpoint: The server address.
        :param kwargs: HTTP headers and other options passed on to the requests
                       module.
        """
        super(HTTPClient, self).__init__(endpoint)
        # Make use of Requests' sessions feature
        self.session = Session()
        self.session.headers.update(self.DEFAULT_HEADERS)
        # Keep last request and response - don't use, will be removed in next
        # major release
        self.last_request = None
        self.last_response = None

    def prepare_request(self, request, **kwargs):
        """
        Prepare the request for sending.

        :param request: The JSON-RPC request (a PreparedRequest object)
        :param kwargs: Configuration for just this request
        :return: None
        """
        # Use the Requests library to prepare the request based on the session
        # configuration
        req = Request(method='POST', url=self.endpoint, data=request, **kwargs)
        request.prepped = self.session.prepare_request(req)
        # Include the http headers in log extra. Will not have effect unless
        # user configures the log format
        request.log_extra = {'http_headers': request.prepped.headers}

    def send_message(self, request, **kwargs):
        """
        Transport the message to the server and return the response.

        :param request: The JSON-RPC request string.
        :param kwargs: Passed on to the requests lib's send function, for last
            minute configuration (Optional)
        :return: The JSON-RPC response.
        :rtype: A string for requests, None for notifications.
        :raise requests.exceptions.RequestException:
            Raised by the requests module in the event of a communications
            error.
        """
        # Keep last request - don't use, will be removed in next major release
        self.last_request = request
        # Send the message with Requests, passing any final config options
        response = self.session.send(request.prepped, **kwargs)
        # Keep last response - don't use, will be removed in next major release
        self.last_response = response
        # Give some extra information to include in the response log entry
        return self.process_response(response.text, log_extra={
            'http_code': response.status_code, 'http_reason': response.reason,
            'http_headers': response.headers}, \
            log_format='<-- %(message)s (%(http_code)s %(http_reason)s)')



from .request import Request, Notification
from .http_client import HTTPClient


def request(endpoint, method, *args, **kwargs):
    """
    A convenience function. Instantiates and executes a HTTPClient request, then
    throws it away.
    """
    return HTTPClient(endpoint).request(method, *args, **kwargs)


def notify(endpoint, method, *args, **kwargs):
    """
    A convenience function. Instantiates and executes a HTTPClient request, then
    throws it away.
    """
    return HTTPClient(endpoint).notify(method, *args, **kwargs)



"""
ZMQ Server.

Deprecated module; use ZeroMQClient instead. Remove in version 3.
"""
from warnings import warn

from .zeromq_client import ZeroMQClient


warn('ZMQServer is deprecated, use ZeroMQClient', DeprecationWarning)

class ZMQServer(ZeroMQClient):
    """Deprecate by subclassing"""
    pass



#/usr/bin/env python
import sys
import pkg_resources

import click

from jsonrpcclient import request, notify
from jsonrpcclient.request import Request, Notification
from jsonrpcclient.http_client import HTTPClient
from jsonrpcclient.exceptions import JsonRpcClientError


version = pkg_resources.require('jsonrpcclient')[0].version


@click.command(context_settings={'ignore_unknown_options': True, 'allow_extra_args': True})
@click.option('--id', default=1, help='Set the id for a request.')
@click.option('--notify', 'request_type', flag_value='notify', help='Indicates that no response is expected.')
@click.option('--send', help='URL to send request to. (requires the Requests library)')
@click.version_option(prog_name='jsonrpcclient', version=version)
@click.argument('method', required=True, metavar='METHOD [PARAMS]...')
@click.pass_context
def main(context, method, request_type, id, send):
    """
    Create a JSON-RPC request.
    """
    exit_status = 0
    # Extract the jsonrpc arguments
    positional = [a for a in context.args if '=' not in a]
    named = {a.split('=')[0]: a.split('=')[1] for a in context.args if '=' in a}
    # Create the request
    if request_type == 'notify':
        req = Notification(method, *positional, **named)
    else:
        req = Request(method, request_id=id, *positional, **named)
    # Sending?
    if send:
        client = HTTPClient(send)
        try:
            response = client.send(req)
        except JsonRpcClientError as e:
            click.echo(str(e), err=True)
            exit_status = 1
        else:
            click.echo(response)
    # Otherwise, simply output the JSON-RPC request.
    else:
        click.echo(str(req))
    sys.exit(exit_status)


if __name__ == '__main__':
    main()



"""
Websockets client.

http://websockets.readthedocs.io/
"""
from .async_client import AsyncClient


class WebSocketsClient(AsyncClient):
    def __init__(self, socket):
        super(WebSocketsClient, self).__init__(None)
        self.socket = socket

    async def send_message(self, request, **kwargs):
        await self.socket.send(request)
        response = await self.socket.recv()
        return self.process_response(response)



"""setup.py"""
from codecs import open as codecs_open
from setuptools import setup

with codecs_open('README.md', 'r', 'utf-8') as f:
    README = f.read()
with codecs_open('HISTORY.md', 'r', 'utf-8') as f:
    HISTORY = f.read()

setup(
    author='Beau Barker',
    author_email='beauinmelbourne@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    description='Send JSON-RPC requests',
    entry_points={'console_scripts': ['jsonrpc = jsonrpcclient.__main__:main']},
    extras_require={
        'aiohttp': ['aiohttp'],
        'requests': ['requests'],
        'requests_security': ['requests[security]'],
        'tornado': ['tornado'],
        'unittest': ['requests', 'pyzmq', 'tornado', 'responses',
                     'testfixtures', 'mock'],
        'websockets': ['websockets'],
        'zmq': ['pyzmq'],
    },
    include_package_data=True,
    install_requires=['future==0.16.0', 'jsonschema==2.6.0', 'click==6.7'],
    license='MIT',
    long_description=README+'\n\n'+HISTORY,
    long_description_content_type='text/markdown',
    name='jsonrpcclient',
    package_data={'jsonrpcclient': ['response-schema.json']},
    packages=['jsonrpcclient'],
    url='https://github.com/bcb/jsonrpcclient',
    version='2.6.0'
)



"""ZeroMQ Client"""
import zmq
import zmq.asyncio

from .async_client import AsyncClient


class ZeroMQAsyncClient(AsyncClient):
    def __init__(self, endpoint, socket_type=zmq.REQ):
        super(ZeroMQAsyncClient, self).__init__(endpoint)
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(socket_type)
        self.socket.connect(endpoint)

    async def send_message(self, request, **kwargs):
        await self.socket.send_multipart((request.encode(),))
        response = await self.socket.recv_multipart()
        return self.process_response(response[0].decode())



"""
aiohttp client.

http://aiohttp.readthedocs.io/
"""
import async_timeout

from .async_client import AsyncClient


class aiohttpClient(AsyncClient):
    """TODO: rename aiohttpClient to AiohttpClient"""
    def __init__(self, session, endpoint):
        super(aiohttpClient, self).__init__(endpoint)
        self.session = session

    async def send_message(self, request):
        with async_timeout.timeout(10):
            async with self.session.post(self.endpoint, data=request) as response:
                response = await response.text()
                return self.process_response(response)
